(load "build/compile.scm")
(load "dclass/emit.scm")

;;; Utilities

(define (butlast xs)
  (define (butlast0 head xs)
    (if (null? (cdr xs)) (reverse head)
	(butlast0 (cons (car xs) head) (cdr xs))))
  (if (null? xs) '()
      (butlast0 '() xs)))

(define (last xs)
  (define (last0 xs)
    (if (null? (cdr xs)) xs
	(last0 (cdr xs))))
  (if (null? xs) '()
      (last0 xs)))

(define (separate by items)
  ;; (separate "," '(1 2 3)) -> (1 "," 2 "," 3)
  (define (separate0 head tail)
    (if (null? tail) (list head)
	(cons head (cons by (separate0 (car tail) (cdr tail))))))
  (if (null? items) items
      (separate0 (car items) (cdr items))))

;;; (gensym name) -> generates a unique name.
;;; (gensym-reset -> resets the gensym counter.
(define gensym #f)
(define gensym-reset #f)
(let ((count 0))
  (set! gensym
	(lambda (name)
	  (let ((it (string->symbol (string-append name count))))
	    (set! count (+ count 1))
	    it)))
  (set! gensym-reset (lambda () (set! count 0))))

;;; Change the following variables:
;;; Root directory of where compiled *.class files should go.
(define dclass-class-base (System.getProperty "user.dir"))
;;; Root directory of where define-class generated *.java files
(define dclass-src-base (System.getProperty "user.dir"))

(define (open-src-file name)
  (let ((f (make-src-file dclass-src-base (class->file (.toString name)))))
    (display (string-append "Writing Java code for " name  " to " f "\n"))
    (.mkdirs (File. (.getParent f)))
    (java.io.PrintWriter. (java.io.BufferedWriter. (java.io.FileWriter. f)))))

(define-macro (define-class . clauses)
  `(define-class-runtime ',clauses))

(define (define-class-runtime clauses)
  (set! cs clauses)
  (let ((class (dclass.Class.)))
    (.clauses$ class (for-each validate-clause clauses))
  (let ((name (full-class-name clauses))) 
    (let ((result (gen-class clauses))
	  (port (open-src-file name)))
      (emit result port)
      (.close port)
      (emit result (current-output-port))
      (display (string-append "Compiling Java code for " name "\n"))
      (compile-class dclass-class-base dclass-src-base name)
      )))

(define (validate-clause clause)
  (let ((clause (group-modifiers clause)))
    clause))



(define (full-class-name clauses)
  (define (package-name clauses)
    (let ((items (collect-clauses '(package) clauses)))
      (if (pair? items) (cadr (car items))
	  "")))
  (string-append (package-name clauses) "." (class-name clauses)))

(define (collect-clauses names clauses)
  (reverse (filter-in (lambda (c) (member (clause-name c) names)) clauses)))

(define find-clause assoc)
(define clause-name car)

(define (group name clauses)
  (let ((it (find-clause name clauses)))
    (if it `(,name ,@(separate "," (cdr it))) '())))

(define (class-name clauses)
  (car (last (cdr (class-clause clauses)))))

(define (class-clause clauses)
  (or (find-clause 'class clauses) (find-clause 'interface clauses)))

(define (check-no-modifiers ms cs)
  (or (null? ms)
      (dclass-clause-error "Clause should not contain modifiers" ms cs)))

(define (check-length= n ms cs)
  (or (= (length cs) n)
      (dclass-clause-error "Clause has wrong number of components" ms cs)))

(define (check-length>= n ms cs)
  (or (>= (length cs) n)
      (dclass-clause-error "Clause has wrong number of components" ms cs)))

(define (dclass-clause-error msg ms cs) (error msg (append ms cs)))

(define (parse-clause clause)
  (if (not (pair? clause)) (error "clause should be a list " clause)
      (let* ((group (group-modifiers clause))
	     (modifiers (car group))
	     (clause (cdr group)))
	(else
	 ((case (car clause)
	    ((class) make-class)
	    ((import) make-import)
	    ((interface) make-class)
	    ((package) make-package)
	    ((property) make-property)
	    ((static) make-static)
	    (else parse-complex-clause)
	    )
	  modifers clause)))))

(define (make-class modifiers clause)
  (check-length>= 2 modifiers clause)
  (let ((cori (car clause))
	(name (cadr clause))
	(clause (cddr clause)))
    (define (make-class0 extends implements)
      (let ((it (dclass.Class.)))
	(.modifiers$ it modifiers)
	(.name$ it name)
	(.extending$ it extends)
	(.implementing$ it implements)
	(.isInterface$ it (eq? cori 'interface))
	it))
    (define (make-extends clause extends implements)
      (cond ((null? clause) (make-class0 extends implements))
	    ((eq? (car clause) 'implements)
	     (make-implements (cdr clause) extends implements))
	    (else (make-extends (cdr clause)
				(cons (car clause) extends)
				implements))))
    (define (make-implements clause extends implements)
      (cond ((null? clause) (make-class0 extends implements))
	    ((eq? (car clause) 'extends)
	     (make-extends (cdr clause) extends implements))
	    (else (make-implements (cdr clause)
				   extends
				   (cons (car clause) implements)))))
    (if (null? clause)
	(make-class0 '() '())
	(cond
	 ((eq? (car clause) 'extends)
	  (make-extends (cdr clause) '() '()))
	 ((eq? (car clause) 'implements)
	  (make-implements (cdr clause) '() '()))))))

(define (invalid-clause modifiers clause)
  (error "Invalid clause " (append modifiers clause)))

(define (make-static modifiers clause)
  (and (check-no-modifiers modifiers clause)
       (dclass.Static. (cdr clause))

(define (make-package modifiers clause)
  (and (check-no-modifiers modifiers clause)
       (check-length= 2 modifiers clause)
       (dclass.Package. (cadr clause)))

(define (make-import modifiers clause)
  (and (check-no-modifiers modiers clause)
       (check-length= 2 modifiers clause)
       (dclass.Import. (cadr clause))))

(define (group-modifiers clause sofar)
  (if (and (pair? clause) (modifier? (car clause)))
      (group-modifiers (cdr clause) (cons (car clause) sofar))
      (cons (reverse sofar) clause)))

(define (make-field modifiers clause)
  (make-field-or-property modifiers clause dclass.Field.))

(define (make-field-or-property modifiers clause constructor)
  ;; clause ::= property type name value?
  (define (make type name . value)
    (let ((it (constructor)))
      (.modifiers$ it modifiers)
      (.type it type)
      (.name$ it name)
      (if (not (null? value)) (.value it (car value)))
      it))
  (apply make clause))

(define (crack-method clause)
  ;; Given:
  ;; (method modifier ... optional(type) (name (type arg) ...) . body)
  ;; Return: method ::= `(,modifiers type ,name ,args ,@body)
  ;; This constructs methods and constructors.
  (define (mods modifiers xs k)
    (if (modifier? (car xs))
	(mods (cons (car xs) modifiers) (cdr xs) k)
	(k (reverse modifiers) xs)))
  (define (type modifiers xs k)
    (if (symbol? (car xs))
	(k (car xs) modifiers (cdr xs))
	(k #f modifiers xs)))
  (define (done type modifiers xs)
	(let ((name (caar xs))
	      (args (cdar xs))
	      (body (cdr xs)))
	  `(,modifiers ,type ,name ,args ,body)))
  (mods '()
	(cdr clause)
	(lambda (modifiers xs) (type modifiers xs done))))

(define (make-method clause)
  (apply (lambda (modifiers type name args body)
	   (dclass.Method. modifiers type name args body))
	 (crack-method clause)))

(define (make-constructor clause)
  (apply (lambda (modifiers type name args body)
	   (dclass.Constructor. modifiers name args body))
	 (crack-method clause)))

(define (gen-class clauses)
  (define (gen-comment clauses)
    `("// Generated by Silk by " ,(System.getProperty "user.name")
      " on " ,(.toString (java.util.Date.)) "\n"))
  (define (gen-package clauses)
    `(package ,(second (find-clause 'package clauses)) ";"))
  (define (gen-imports clauses)
    (append 
     (map (lambda (i) `(import ,i ";"))
	  (apply append (map cdr (collect-clauses '(import) clauses))))
     '((import silk.Procedure ";")
       (import silk.SI ";"))))
  (define (gen-class-line clauses)
    (let* ((class (class-clause clauses))
	   (modifiers (butlast (cdr class)))
	   (name (class-name clauses))
	   (type (clause-name class))
	   (extends (group 'extends clauses))
	   (implements (group 'implements clauses)))
      `(,@modifiers ,type ,name ,@extends ,@implements)))
  (gensym-reset)
  `(,(gen-comment clauses)
    ,(gen-package clauses)
    ,@(gen-imports clauses)
    ,(gen-class-line clauses)
    { ,@(gen-class-body clauses) } ))

(define (gen-class-body clauses)
  (map gen-class-member
       (collect-clauses '(constructor field method property static) clauses)))

(define (gen-class-member clause)
  ((case (car clause)
     ((field) gen-field)
     ((method) gen-method)
     ((constructor) gen-constructor)
     ((property) gen-property)
     ((static) gen-static))
   clause))


(define (gen-field field)
  (let ((f (make-field field)))
    (cond
     ((null? (dclass-value f))
      `(,@(dclass-modifiers f) ,(dclass-type f) ,(dclass-name f) ";"))
     ((dclass-constant? (car (dclass-value f)))
      `(,@(dclass-modifiers f) ,(dclass-type f) ,(dclass-name f)
	= ,(dclass-java-value (car (dclass-value f))) ";"))
     (else
      (let* ((v (car (dclass-value f)))
	     (symbol (dclass-member-name "F_" f))
	     (proc (dclass-procedure
		    symbol (if (dclass-static? f) '() '(this)) (list v)))
	     (value (dclass-call (if (dclass-static? f) (list symbol)
				     (list symbol 'this))
				 (dclass-type f))))
	(list 
	 proc
	 `(,@(dclass-modifiers f) ,(dclass-type f) ,(dclass-name f)
	   = ,value ";")))))))
	  
(define dclass-static-body cdr)
(define dclass-modifiers car)
(define (dclass-static? f) (member 'static (dclass-modifiers f)))
(define dclass-type cadr)
(define dclass-name caddr)
(define dclass-value cadddr)
(define dclass-args cadddr)
(define dclass-body (lambda (x) (cdr (cdddr x))))
(define (modifier? x) (member x '(
				  final
				  private
				  protected
				  public
				  static
				  strictfp
				  synchronized
				  transient
				  volatile
				  )))

(define (dclass-constant? x)
  ;; Is x a Scheme constant that is easy to convert to Java?
  (or (number? x) (string? x) (boolean? x)))

(define (dclass-java-value x)
  ;; Turn a Scheme constant into a string representing the Java value.
  (if (eq? x #null) 'null
      (dclass-java-value0 x)))
(define-method (dclass-java-value0 (x Boolean)) (if x "true" "false"))
(define-method (dclass-java-value0 (x Character)) (string-append #\' x #\'))
(define-method (dclass-java-value0 (x Float)) (string-append x "F"))
(define-method (dclass-java-value0 (x Long)) (string-append x "L"))
(define-method (dclass-java-value0 (x Number)) (.toString x))
(define-method (dclass-java-value0 (x Object))
  (error "Can't convert " x " to a Java value."))
(define-method (dclass-java-value0 (x String))
  (string-append "\"" (quotify x) "\""))

(define (quotify x)
  (define (find x n)
    (let ((end (.indexOf x "\"" n)))
      (if (= end -1) (.length x) end)))
  (define (quotify0 x start end so-far)
    (if (< end (.length x))
	(quotify0 x end (find x (+ end 1))
		  (cons "\\" (cons (.substring x start end) so-far)))
	(apply string-append (reverse (cons (.substring x start end)
					    so-far)))))
    (quotify0 x 0 (find x 0) '()))

(define (dclass-procedure symbol args body)
  (let ((p `(lambda ,args ,@body)))
    `(private static final Procedure ,symbol = "(" Procedure ")" 
	     SI.eval "(" ,(dclass-java-value (.toString p)) ")" ";")))

(define (test-stringify2)
  (let ((value '(string-append "x" "-" "y")))
    (dclass-java-value (.toString `(lambda () ,value)))))

(define (dclass-call args type)
  (let ((type (if (pair? type) (car type) type)))
    (if (or (eq? type 'void) (null? type))
	`("(" SI.call "(" ,@(separate "," args) ")" ")")
	`(,(silk->java type) "(" SI.call "(" ,@(separate "," args) ")" ")"))))

(define (silk->java type)
  (let ((it (assoc type '(
			  (boolean SI.booleanValue)
			  (byte SI.byteValue)
			  (char SI.charValue)
			  (short SI.shortValue)
			  (int SI.intValue)
			  (long SI.longValue)
			  (float SI.floatValue)
			  (double SI.doubleValue)))))
    (if it (cadr it) (string-append "(" type ")"))))

(define (dclass-member-name prefix m)
  ;; Don't upcase name which might make it nonunique.
  (string->symbol
   (string-append "DCLASS_"
		  (if (equal? prefix "C_") ; Constructor.
		      "C_"
		      (string-append prefix (if (dclass-static? m) "S_" "I_")))
		  (.toString (dclass-name m)))))

(define (gen-method clause)
  (define (gen-method m)
    ;; Need to handle this and super.
    (define (needs-return? type)
      (not (or (null? type) (eq? (car type) 'void)))) 
    (let* ((args (map cadr (dclass-args m)))
	   (args (cons (dclass-member-name "M_" m)
		       (if (dclass-static? m)
			   args
			   (cons 'this args)))))
      `(,@(dclass-modifiers m) ,@(dclass-type m)
	,(dclass-name m)
	"(" ,(separate "," (dclass-args m)) ")" "{"
	,@(if (needs-return? (dclass-type m)) '(return) '())
	,(dclass-call args (dclass-type m))
	";" "}")))
		 
  (define (gen-static m)
    (dclass-procedure (dclass-member-name "M_" m)
		      (if (dclass-static? m)
			  (map cadr (dclass-args m))
			  (cons 'this (map cadr (dclass-args m))))
		      (dclass-body m)))

  (let ((m (make-method clause)))
    (list
     (gen-method m)
     (gen-static m))))

(define (gen-property clause)
  ;; +++ Need set and get methods too!
  (gen-field clause))

(define (gen-static clause)
  (let ((symbol (gensym "DCLASS_STATIC")))
    (list
     (dclass-procedure symbol '() (dclass-static-body clause))
     `(static "{"
	      (dclass-call (list symbol) 'void)
	      "}"))))

;;; Constructors and methods are the same except for the first form in
;;; their body.  Constructors can contain (this ...) or (super ...).
;;; The first form in a constructor body is called the "head".

(define (gen-constructor clause)
  (define (body-head? m)
    (let ((body  (dclass-body m)))
      (and (not (null? body))
	   (pair? (car body))
	   (member (caar body) '(this super)))))
  (define (body-head m)
    (and (body-head? m) (car (dclass-body m))))
  (define (body-tail m)
    (if (body-head? m) (cdr (dclass-body m))
	(dclass-body m)))
  (define (gen-method m)
    ;; +++ Need to handle this and super.
    (let* ((args (map cadr (dclass-args m)))
	   (args (cons (dclass-member-name "C_" m)
		       (cons 'this args)))
	   (head (body-head m))
	   (tail (body-tail m)))
      `(,@(dclass-modifiers m) ,@(dclass-type m)
	,(dclass-name m)
	"(" ,(separate "," (dclass-args m)) ")" "{"
	,(if head (gen-constructor-head head) '())
	,(if (null? tail) '()
	     `(,(dclass-call args (dclass-type m)) ";"))
	"}")))
  (define (gen-static m)
    (let ((tail (body-tail m)))
      (if (null? tail) '()
	  (dclass-procedure (dclass-member-name "C_" m)
			    (cons 'this (map cadr (dclass-args m)))
			    tail))))
  (let ((m (make-constructor clause)))
    (list
     (gen-method m)
     (gen-static m))))



'(assert
 (equal?
  (body-head
  (make-constructor
   '(constructor
     public (Compare2 (Procedure predicate))
     (super predicate))))
  '(super predicate)))

(define (gen-constructor-head head)
  (define (head-simple? head)
    (every   (lambda (x) (or (symbol? x) (not (pair? x)))) head))
  (if (head-simple? head)
      `(,(car head) "(" ,(separate "," (cdr head)) ")" ";")
      (error "Can't handle complex head " head)))

(define (test-2)
  (define-class
    (package frog)
    (import java.util.Comparator)
    (public class Compare implements Comparator)
    ;; Design issue, fields must be public for Silk code to access.

    (public Procedure predicate)	

    (public Compare (Procedure predicate)
     (.predicate$ this predicate))

    (public boolean predicate (Object a Object b)
     ((.predicate$ this) a b))

    (public int compare (Object  Object b)
     (cond ((.predicate this a b) -1)
	   ((.predicate this b a) 1)
	   (else 0)))

    (public boolean equals (Object that)
     (and (eq? (.getClass this) (.getClass that))
	  (eq? (.predicate$ this) (.predicate$ that))))

    (public int (hashCode) 0))
  (let ((a #(1 5 2 4 9 8 3 7 6 2 4)))
    (print a)
    ;; (java.util.Arrays.sort a (frog.Compare. (lambda (a b) (not (= a b)))))
    ;; (print a)
    (java.util.Arrays.sort a (frog.Compare. >))
    (print a)
    (java.util.Arrays.sort a (frog.Compare. <))
    (print a)
    #t))


;;;KRA 28JAN02: Currently super not handled.
(define (test-3)
  (define-class
    (package frog)
    (import java.util.Comparator)
    (class public Compare2 extends Compare)

    (property public int predicateCount 0)
    (property public int compareCount 0)
    (public Compare2 (Procedure predicate)
     (super predicate))
    (public boolean predicate (Object a Object b)
     (.predicateCount$ this (+ (.predicateCount$ this) 1))
     (.predicate super a b))
    (public int compare (Object a Object b)
     (.compareCount$ this (+ (.compareCount$ this) 1))
     (.compare super a b)))
  (let ((a #(1 5 2 4 9 8 3 7 6 2 4)))
    (print a)
    (java.util.Arrays.sort a (frog.Compare2. >))
    (print a)
    (print `(predicate: ,(.getPreciateCount a)
	     compare:   ,(.getCompareCount a)))))

(define (decode-reflector name)
  (define (trim name)
    (define (trim-start name) (if (.startsWith name ".") 1 0))
    (define (trim-end name)
      (if (or (.endsWith name "$") (.endsWith name ".")) (- (.length name) 1)
	  (.length name)))
    (.substring name (trim-start name) (trim-end name)))

  (let* ((name (.toString name))
	 (iors (if (.startsWith name ".") 'instance 'static))
	 (m (if (.endsWith name "$") 'field
		(if (.endsWith name ".") 'constructor
		    'method))))
    (list (trim name) iors m)))
