// Copyright 2001, Jonathan Bachrach.  See file TERMS.

import java.io.*;
import java.util.*;
import java.lang.reflect.*;
import antlr.*;

/**
 * Mother of all Fragments
 * @author  Jonathan Bachrach
 * @author  Keith Playford
 * @version 0.9, 03OCT00
 */
public abstract class Fragment implements Cloneable {
    /** Stream for pretty printing using pprint and pprintln */
    public static PrettyStream out = new PrettyStream(System.out);
    StringFragment substituteAsString () {
	return new StringFragment(stringify());
    }
    /** Top level entry point to compiling fragments within a pattern
     ** @see     Template#compilePattern
     ** @param   tail    remaining fragments to compiling through tail calling
     ** @param   nowVar  runtime variable bound to current input fragments
     ** @returns pattern matcher code for this fragment and tail fragments
     **/
    public Expansion compilePatternElement
	(FragmentList tail, IdentifierFragment nowVar, IdentifierFragment endVar) {
	// #{ ?m-code; ?m*-code } package macros;

	IdentifierFragment nextVar = IdentifierFragment.genSym("fs");
	Expansion mCode            = generatePatternCode(nowVar, nextVar);
	Expansion mTailCode        = tail.compilePatternElements(nextVar, endVar);
	return Template.processTemplate
	    (FragmentList.fnil
	     .fpush(mCode)
	     .fpush(mTailCode)
	     .freverse());
    }
    /** Pattern matching code generator 
     ** @see     Fragment#compilePatternElement
     ** @param   nowVar  runtime variable bound to current input fragments
     ** @param   nextVar where to put remaining fragments after matching
     ** @returns pattern matcher code for just this fragment
     **/
    public Expansion generatePatternCode
	(IdentifierFragment nowVar, IdentifierFragment nextVar) {
	System.out.println("NYI gpc " + this.getClass() + " " + this);
	return new Expansion(FragmentList.fnil);
    } 
    /** Template code generator -- probably should be compileTemplateElement
     ** @see     Template#compileTemplate
     ** @returns template generator code for this fragment
     **/
    public Fragment compileTemplate(int depth) {
	System.out.println("NYI ct " + this);
	return this;
    } 
    public Fragment compileQuestionExpression (int tDepth, int qDepth) {
	System.out.println("NYI ctv " + this);
	return this;
    }
    /** @returns list of all pattern variables within fragment
     **/
    public FragmentList collectBoundVariables () { return FragmentList.fnil; }
    /** Used to record line of macro use for artificially constructed tokens 
     **/
    private static ThreadLocal _currentLine = new ThreadLocal();
    public static int getCurrentLine() { 
	return ((Integer)_currentLine.get()).intValue();
    }
    public static void setCurrentLine(int line) { 
	_currentLine.set(new Integer(line));
    }
    public int getLine() { 
	Token t = (Token)tokens().getHead();
	return t.getLine();
    }
    /** @returns Lexer token used for generating artificial tokens 
     **/
    public static Token makeToken (int type, String text) {
	Token t = new CommonToken(type, text);
	if (_currentLine.get() != null)
	    t.setLine(getCurrentLine());
	return t;
    }
    public static Token makeToken (int type) {
	return makeToken(type, "");
    }
    /** @returns remembered tokens that were present during original lexing
     **/
    public List tokens () {
	System.out.println("NYI ct " + this.getClass());
	return List.nil;
    } 
    /** @returns fragment with all codeQuotes compiled as templates
     **/
    public Fragment expandTemplates () { 
	return this;
    } 
    /** @returns fragment with one level of macro's expanded
     **/
    public Fragment expandMacros (MacroExpander expander, boolean isRecursive) { 
	return this;
    } 
    /** @returns macro calls converted to MacroCall fragments
     **/
    public Fragment markMacros (MacroExpander expander) { 
	return this;
    } 
    /** @returns does this fragment represent the end of a clause
     **/
    public boolean isTerminator () {
	return false;
    }
    /** @param name  string to be capitalized
	@returns name with first letter uppercased
     **/
    public static String capitalize (String name) {
	return Character.toUpperCase
	    (name.charAt(0)) + name.substring(1).toLowerCase();
    }

    // value access protocol 

    /** protocol for converting fragments back into Java literals
     ** @returns internal value either literal or a sequence of one literal
     **/
    public Object getValue () { return this; }
    public int getInt () { return ((Integer)getValue()).intValue(); }
    public float getFloat () { return ((Float)getValue()).floatValue(); }
    public String getString () { return ((String)getValue()); }
    public char getChar () { return ((Character)getValue()).charValue(); }
    public abstract FragmentList getInsideFragments();

    // printing protocol

    void print(PrettyStream p) {p.print(this);}
    void print() {print(Fragment.out);}

    void println() {println(Fragment.out);}
    void println(PrettyStream p) {print(); p.println();}

    // pretty printing protocol

    /** pretty print an object at a particular depth
     ** @param p      stream
     ** @param depth  indentation
     ** @param isCode how to treat variable references true is unadorned
     **/
    void pprint() {
	pprint(Fragment.out, 0, false);
    }
    void pprint(int depth) {
	pprint(Fragment.out, depth, false);
    }
    void pprint(int depth, boolean isCode) {
	pprint(Fragment.out, depth, isCode);
    }
    void pprint(PrettyStream p, int depth, boolean isCode) { 
	p.print(toString()); 
    }
    // pretty print an object with newline
    private static int indentation = 2;
    void pprintln(PrettyStream p, int depth) { 
	// p.print("D" + depth);
	p.println();
	for (int i = 0; i < depth; i++) 
	    for (int j = 0; j < indentation; j++)
		p.print(" ");
    }
    // pretty print an element of a sequence 
    void pprinte(PrettyStream p, int depth, boolean isLast, boolean isCode) { 
	pprint(p, depth, isCode);
	if (!isLast && !p.atBeginning) // need for separator?
	    p.print(" ");
    }
    ///
    /// LPRINT -- LINE PRINTING
    void lprint(int depth) {
	lprint(Fragment.out, depth, false);
    }
    void lprint(int depth, boolean isCode) {
	lprint(Fragment.out, depth, isCode);
    }
    void lseek(PrettyStream p, int line, int depth) {
	/*
	System.out.print("CUR " + p.getLine() + " SEEKING " + line);
	if (line == 0)
	    System.out.println(" " + this);
	else
	    System.out.println();
	*/
	while (p.getLine() < line) {
	    p.println();
	    for (int i = 0; i < depth; i++) 
		for (int j = 0; j < indentation; j++)
		    p.print(" ");
	}
    }
    void lseek(PrettyStream p, int depth) {
	lseek(p, getLine(), depth);
    }
    void lprint(PrettyStream p, int depth, boolean isCode) { 
	lseek(p, depth);
	pprint(p, depth, isCode); 
    }
    void lprint(PrettyStream p, String s, int line, int depth) { 
	lseek(p, line, depth);
	p.print(s); 
    }
    // pretty print an element of a sequence 
    void lprinte(PrettyStream p, int depth, boolean isLast, boolean isCode) { 
	lprint(p, depth, isCode);
	if (!isLast && !p.atBeginning) // need for separator?
	    p.print(" ");
    }
    String stringify() { return toString(); }
    // death method
    // TODO: should throw deadly exception
    public static void die (String message) {
	System.err.println("Deadly error: " + message);
    }
}

