;;;
;;; Application specific Build stuff.
;;;

(load "build/compile.scm")
(load "elf/sort.silk")
(load "elf/classpath.scm")
(extendClassPath "SRC")

(load "elf/basic.scm")
(import "java.io.File")

(define s+ string-append)
(define (path . args)
  ;; Convert arguments into a CLASSPATH or PATH string.
  (apply s+ (separate ($ "path.separator") (flatten args))))

(define (maybe-jar f)
  (or (.endsWith (.getName f) ".jar") (.endsWith (.getName f) ".zip")))

(define jar-files
  ;; Collect the .jar files in directory.
  (lambda (directory) (reverse (filter-in maybe-jar (.listFiles directory)))))

(define (app-home)
  ;; Where the home directory of the application is on this machine.
  (let ((it (System.getProperty "app.home")))
    (if (eq? it #null) (System.getProperty "user.dir")
	it)))

(define (file root . path)
  (define (file0 root path)
    (if (null? path) root
	(file0 (File. root (car path)) (cdr path))))
  (file0 root path))

;;;
;;; Application specific bootstrap code.
;;;
(define (src-dir) (file (app-home) "SRC"))
(define (class-dir) (file (app-home) "SRC"))

(define (jar-roots)
  ;; List of directories containing .jars.
  (list (file (app-home) "lib")
	(file (app-home) "antlr-2.6.0")))

(define (class-path)
  ;; List the .jar files needed for this application.
  (flatten (list
	    (src-dir)			; may contain .scm or other files.
	    (file (src-dir) "EXAMPLES")
	    (file (src-dir) "TESTS")
	    (map jar-files (jar-roots))
	    )))

(define (D)
  ;; List of -D argument strings.
  (list (string-append "-Dapp.home=" (app-home))
	))

(replaceClassPath (class-path))
(load "build/builder.scm")

(define (jse-file? f) (.endsWith (.getName f) ".jse"))

(define (change-extension f type)
  (let ((f (.toString f)))
    (File. (string-append (.substring f 0 (.lastIndexOf f ".")) type))))

(define (src-file f) (file (src-dir) f))

(define (javac-options) '(-g))

(define *commands* '())
(define (commands) (reverse *commands*))
(define (add-command name documentation procedure)
  ;; Add a command to the list of commands.
  (set! *commands* (cons (list name documentation procedure) *commands*)))

(add-command '-help
	     "-help             = Print this message."
	     cmd-help)

(define (cmd-clean)
  ;; Delete .class files.
  (delete-class-files (src-dir))
  ;; Delete .java files produced from .jse files.
  (for-each
   (lambda (f) (.delete (change-extension f ".java")))
   (all-files (src-dir) jse-file?))
  )

(add-command '-clean
	     "-clean            = Start with a clean slate."
	     cmd-clean)

(define antlr-output
  '("JavaLexer.java"
    "JavaRecognizer.java"
    "JavaTokenTypes.java"))

(define (update-antlr)
  ;; update the Java parser.
  (let ((targets  (map src-file antlr-output))
	(sources (list (src-file "java.g"))))
    (if (needs-update? targets sources)
	(begin 
	  (run `(java -cp ,(file (app-home) "antlr-2.6.0" "antlr-tool.jar")
		      antlr.Tool
		      -o ,(src-dir)
		      ,@sources))
	  #t)
	#f)))

(define (jse->class-action f)
  ;; Update a .class file from a jse file if necessary.
  (if (needs-update? (list (change-extension f ".java")) (list f))
      (begin
	(print `(jse ,f))
	(run `(java -classpath ,(path (class-path)) JSE ,f))
	(java->class-action (change-extension f ".java")))))

(define (java->class-action f)
  ;; Update a .class file from a .java file if necessary.
  (if (needs-update? (list (change-extension f ".class")) (list f))
      (begin
	(print `(javac ,f))
	(run `(javac -classpath ,(path (class-path)) ,(javac-options) ,f)))))

(define (update-jse files)
  ;; Update a group if .jse files.
  (for-each jse->class-action files))
(define (update-classes java-files)
  ;; Update a group of .java files.
  (for-each java->class-action java-files))

(define (syntax-files)
  ;; These files must be compiled in order.
  '("syntaxSwitchBodySConstraint.jse"
    "syntaxSExpander.jse"
    "ruleSExpander.jse"
    "auxRuleSExpander.jse"))

(define (update-all)
  (update-antlr)
  (update-classes (java-files (src-dir)))
  (update-jse (map src-file (syntax-files)))
  ;; At this point JSE is up!
  (update-jse (filter-in jse-file? (.listFiles (file (src-dir) "EXAMPLES")))))

(define (cmd-compile) (update-all))

(add-command '-compile
	     "-compile          = Recompile updated sources."
	     cmd-compile)

(define (cmd-jse . args) (JSE.main (list->array String.class args)))

(add-command '-jse
	     "-jse arg*         = Run JSE with the given arguments."
	     cmd-jse)

