;;;
;;; Build Framework.
;;;

;;; This should be application independent.  Applications should
;;; override procedures, as needed, after loading this file.

(define $ System.getProperty)

;;; Overrides the one in build\compile.scm.
(define (mkdir dir)
  (if (not (.exists dir))
      (begin (print `(making directory ,dir))
	     (.mkdirs dir)))
  dir)

(define (all-directories dir) (directory-walk dir cons '()))

(define (directory-tree-walk directory how so-far)
  ;;; Walk files directly under directory.
  ;; KRA 27DEC01: directory-walk and file-walk should be written in
  ;; terms of this.
  (foldL (.listFiles directory) how so-far))

(define (delete-files-under directory)
  ;; Delete the file hierarchy under directory.
  (define (delete f)
    (let ((result (.delete f)))
      (brl-display "[result] [f]\n")
      result))
  (define (how f undeleted)
    (if (.isDirectory f)
	(let ((undeleted (directory-tree-walk f how undeleted)))
	  (if (delete f) undeleted
	      (cons f undeleted)))
	(if (delete f) undeleted
	    (cons f undeleted))))
  (brl-display "Deleting files under [directory].\n")
  (directory-tree-walk directory how '()))

(define (delete-files-under-and-warn directory)
  (define (warn undeleted)
    (brl-display "Warning: While deleting [directory] the following files could not be deleted:\n")
    (for-each print undeleted))
  (let ((undeleted (delete-files-under directory)))
    (if (not (null? undeleted)) (warn undeleted))
    undeleted))

(define (html-file? f) (.endsWith (.getName f) ".html"))

(define (compile-class class-base src-base name)
  (and (compile-file class-base src-base (class->file name))
       (let ((c (class name)))
	 ;; Sometimes a java file can have the wrong package defined
	 (if (eq? c #null) (print (list "Can't find class" name))
	     (if (remote? c)
		 (begin
		   (display "rmicing ") (display name) (newline)
		   (rmic "-classpath" (System.getProperty "java.class.path")
			 "-d" class-base
			 ;; "-sourcepath" src-base
			 name))))
	 #t)))

;;; from silkweb/build/compile.scm
(define (recompile class-base src-base package recurse?)
  ;; Recompile an entire package, and subpackages if recurse? is #t.
  ;; Try to recompile everything that needs it, but return #t
  ;; only if there are no errors.
  (display "Recompiling package ") (display package)
  (if recurse? (display " recursively"))
  (newline)
  ;; Make the class directory if necessary.
  (.mkdirs (File. (package->file class-base package)))
  (let ((directory (File. (package->file src-base package))))
    (recompile-files class-base src-base (java-files-no-aspects directory))
    (if recurse?
	(map* (lambda (f)
		(recompile class-base src-base 
			   (file->package src-base f)
			   recurse?))
	      (directories directory)))))

;;; command ::= `(,name ,documentation , procedure)
;;; name ::= Symbol strarting with "-", ie -cvs.
;;; documentation ::= Documentation String, can be multiple lines.
;;; procedure ::= a procedure applied to its arguments when command is invoked.

(define (commands) '())
(define-method (find-command (name Symbol)) (assoc name (commands)))
(define-method (find-command (name String))
  (find-command (string->symbol name)))

(define command-keyword? find-command)
(define command-name car)
(define command-documentation cadr)
(define command-procedure caddr)

;;; Do something on startup.  Override this!
(define (builder-init) #f)

(define (builder args)
  (builder-init)
  (if (or (eq? args #null) (= (vector-length args) 0))
      (builder-frame)
      (begin
	(builder-args (vector->list args))
	(if builder-exit? (builder-exit))
	)))

(define (builder-frame) (silk.Scheme.readEvalWriteLoop "> "))

;;; Should builder exit after processing all commands?
(define builder-exit? #t)		

(define (builder-exit)
  (print `(builder exiting))
  (for-each
   (lambda (p)
     (print `(destorying ,p))
     (.destroy p))
   processes)
  (System.exit 0))

(define processes '())

(define cmd-help
  (let ((print? #t))
    (lambda ()
      (if print?
	  (begin
	    (display "build command*

With no commands an interactive prompt is provided.
command ::= -key arg*

Where command is one of:

")
	    (for-each (lambda (c)
			(display (command-documentation c))
			(newline))
		      (commands))
	    (set! print? #f))))))

(define (builder-args args)
  ;; Process args and invoke commands.
  (cond
   ((null? args) 'done)
   ((command-keyword? (car args))
    (builder-args-keyword (find-command (car args)) (cdr args) '()))
   ((.startsWith (car args) "(")
    (silk.SI.eval (car args))
    (builder-args (cdr args)))
   ((not (.startsWith (car args) "-"))
    (load (car args))
    (builder-args (cdr args)))
   (else (display+ "Unknown argument: " (car args) "\n")
	 (cmd-help))
    ))

(define (builder-args-keyword command args so-far)
  ;; Accumulate the list of args for this command in so-far.
  ;; Apply the command to its arguments and continue processing.
  (cond
   ((or (null? args) (command-keyword? (car args)))
    (command-apply command (reverse so-far))
    (builder-args args))
   (else (builder-args-keyword command
			       (cdr args)
			       (cons (car args) so-far)))))

(define (command-apply command args)
  (print (time (begin
		 (apply (command-procedure command) args)
		 (command-name command))
	       1)))

(define (command-run name . args)
  (command-apply (find-command name) args))

;;; Utilities.
(define (display+ . args) (display (apply string-append args)))
